/*
 *	$Id: calc_server.c,v 1.9 2005/04/28 22:25:40 rh Exp $
 *
 * calc_server: a simple RPN calculating server
 *
 */

#define	_POSIX_C_SOURCE	200112	/* required for fdopen() */
#define __EXTENSIONS__		/* required on Solaris */

#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "calculator.h"

#define BACKLOG 4		/* number of pending connections */

#ifdef BSD			/* is this a BSD system? */
#define	ON_BSD(code)	code	/* then compile BSD specific code */
#else
#define	ON_BSD(code)		/* on other systems, just ignore this */

#if !defined(socklen_t) && !defined(__socklen_t_defined)
typedef int socklen_t;		/* define our own socklen_t if not defined */
#endif

#endif /* if/else BSD */

/*
 * a simple calculator child that handles one incoming connection
 * before invoking the actual calculator, it creates a simple
 * buffered I/O environment to allow simple stdio type operations
 *
 * parameters:
 *	- sin: the TCP socket representing the current incoming connection
 *
 * return values:
 *	this never returns, but exits with EXIT_SUCCESS or
 *	EXIT_FAILURE
 */
void calc_child(int sin)
{
	FILE *in, *out;		/* stdio file pointers */
	int sout = dup(sin);	/* duplicate the socket for output */

	if (sout == -1)
	{
		perror("cannot duplicate socket");
		exit(EXIT_FAILURE);
	}

	in = fdopen(sin, "r");		/* create a buffered reader */

	if (in == NULL)
	{
		perror("cannot create buffered reader");
		exit(EXIT_FAILURE);
	}

	out = fdopen(sout, "w");	/* create a buffered writer */

	if (out == NULL)
	{
		perror("cannot create buffered writer");
		exit(EXIT_FAILURE);
	}

	/* run the calculator and exit */

	exit(calculator(in, out) ? EXIT_SUCCESS : EXIT_FAILURE);
}


/*
 * create a server TCP socked bound to a certain port
 *
 * parameters:
 *	- port: the port to bind to
 *
 * return values:
 *	-1 in case of failure
 *	a valid socket >= 0 otherwise
 */
int server_socket(int port)
{
	struct sockaddr_in local;	/* the local address to bind to */
	int sock;			/* the socket to be created */

	memset(&local, 0, sizeof(local));	/* clear address */

	ON_BSD(local.sin_len = sizeof(local));
	local.sin_family = AF_INET;		/* internet socket */
	local.sin_port = htons(port);		/* port to bind to */

	sock = socket(PF_INET, SOCK_STREAM, 0);	/* get me a socket */

	if (sock < 0)				/* error? */
	{
		perror("cannot create socket");
		return -1;
	}

	/* bind socket */
	if (bind(sock, (struct sockaddr *) &local, sizeof(local)) != 0)
	{
		fprintf(stderr, "cannot bind socket to port %d: %s\n",
			port, strerror(errno));
		return -1;			/* error? */
	}

	if (listen(sock, BACKLOG) != 0)		/* listen on port */
	{
		fprintf(stderr, "cannot listen on port %d: %s\n",
			port, strerror(errno));
		return -1;
	}

	return sock;				/* success! */
}

/*
 * cleanup function: join with all children
 *
 * parameters:
 *	none
 *
 * return values:
 *	none
 */
void signal_handler(int sig)
{
	int status;

	while (waitpid(-1, &status, 0) >= 0)	/* collect children */
		;

	exit(EXIT_SUCCESS);
}


/*
 * fork off a child handling a connection on a given socket
 *
 * parameters:
 *	- conn: connection socket
 *
 * return values:
 *	 0 if successful
 *	-1 in case of failure
 */
int fork_child(int conn)
{
	pid_t pid = fork();			/* create a child */

	if (pid == -1)				/* error? */
	{
		perror("cannot fork");
		return -1;
	}

	if (pid == 0) calc_child(conn);		/* child */ 

	/*
	 * calc_child() never returns, so everything below
	 * is the cleanup required by the parent.  The parent
	 * does not do anything with the incoming connection,
	 * so we simply close the connection from the parent side
	 */

	return close(conn);
}


/*
 * accept incoming connections on the given (bound) socket and
 * spawn children accordingly
 *
 * parameters:
 *	- sock: the bound TCP (server) socket
 *
 * return values:
 *	EXIT_FAILURE in case of failure
 */
int accept_connections(int sock)
{
	struct sockaddr_in incoming;		/* incoming connection info */
	socklen_t size = sizeof(incoming);	/* size of that info */
	int conn;				/* connection socket */

	for(;;)					/* forever */
	{
		memset(&incoming, 0, size);

		ON_BSD(incoming.sin_len = size);

		conn = accept(sock, (struct sockaddr *)&incoming, &size);

		if (conn == -1) break;		/* error */

		if (fork_child(conn)) break;	/* child handles connection */
	}

	return EXIT_FAILURE;
}


/*
 * wait for incoming connections on a given port and spawn off
 * child processes for each incoming connection
 *
 * parameters:
 *	- port: the port to bind to
 *
 * return values:
 *	EXIT_FAILURE in case of failure, won't return otherwise
 */
int calc_server(int port)
{
	int sock = server_socket(port);		/* get a network socket */

	if (sock == -1) return EXIT_FAILURE;

	/*
	 * invoke signal_handler() on Control-C
	 */
	if (signal(SIGINT, signal_handler) == SIG_ERR)
		perror("cannot install signal handler");

	return accept_connections(sock);	/* accept connections */
}

