/*
 *	$Id: calculator.c,v 1.8 2005/04/28 21:05:14 rh Exp $
 *
 * calculator: a simple RPN calculator
 *
 */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>

#include "calculator.h"

#ifndef INFINITY
#define INFINITY HUGE_VAL
#endif

/*
 * perform a basic calculation operation, depending on the input
 * if the input is a number, push the number onto the stack
 * otherwise, try to interpret the command as an operation and
 * perform the operation
 *
 * parameters:
 *	- stack: the stack to operate on
 *	- string: the input string
 *	- out: the output file handle to print results to (e.g. stdout)
 *
 * return values:
 *	1 if successful
 *	0 otherwise
 */
int calculate(Stack *stack, char *string, FILE *out)
{
	double left, right;	/* the two values for an operation */
	double result;		/* result */
	char op;		/* the operation */
	int unary;		/* unary or binary operation? */

	if (isdigit((int)*string))	/* a number? -> push onto stack */
	{
		if (stack_pushDouble(stack, atof(string)))
			return 1;	/* value pushed successfully */

		fprintf(out, "cannot push %s onto stack: %s\n",
				string, strerror(errno));
		return 0;
	}

	/* not a number, so perform an operation */

	op = tolower((int)*string);

	if (stack_isEmpty(stack))
	{
		fprintf(out, "empty stack.\n");
		return 0;
	}

	right  = stack_popDouble(stack);

	unary = 1;
	switch(op)		/* unary operation? */
	{
		case 'a': result = fabs(right); break;
		case 'c': result = ceil(right); break;
		case 'f': result = floor(right); break;
		case 'e': result = exp(right); break;
		case 'l': result = log(right); break;
		case 'p': fprintf(out, "%g\n", result = right); break;
		case 's': result = sqrt(right); break;
		default: unary = 0;
	}

	if (!unary)
	{
		if (stack_isEmpty(stack))
		{
			stack_pushDouble(stack, right);
			fprintf(out, "Cannot perform '%s': stack underflow.\n",
				string);
			return 0;
		}

		left = stack_popDouble(stack);

		switch (op)
		{
			case '+': result = left + right; break;
			case '-': result = left - right; break;
			case '*': result = left * right; break;
			case '/': result = left / right; break;
			case '%': result = fmod(left, right); break;
			case '^': result = pow(left, right); break;
			default:
				fprintf(out, "Unknown operation '%s'\n",
					string);
				return 0;
		}
	}

	if (!stack_pushDouble(stack, result))
	{
		fprintf(out, "Cannot push result %g onto stack: %s\n", result,
			strerror(errno));
		return 0;
	}

	return 1;	/* success */
}


/*
 * a simple calculator that reads from stdio and performs stack
 * based calculations
 *
 * parameters:
 *	- in:  the input stream (e.g. stdin)
 *	- out: the output stream (e.g. stdout)
 *
 * return values:
 *	0 on success
 *	1 on failure
 */
int calculator(FILE *in, FILE *out)
{
	char input[81];
	Stack *stack = new_Stack();

	if (stack == NULL)
	{
		fprintf(out, "cannot allocate stack: %s\n", strerror(errno));
		return EXIT_FAILURE;
	}

	while(fscanf(in, "%80s", input) > 0)
	{
		if (tolower((int)*input) == 'x')
			break;

		calculate(stack, input, out);
		fflush(out);
	}

	stack_Release(stack);

	if (!ferror(in) && !ferror(out)) return EXIT_SUCCESS;

	perror("I/O error");

	return EXIT_FAILURE;
}


#ifdef TEST
int main(int argc, char *argv[])
{
	return calculator(stdin, stdout);
}
#endif
